---------------------------------------------------------------
-- NON RESTORING Division algorithm for integer (signed) X and Y
-- NBITS >= MBITS
-- Returns a Quotient Q of NBITS and a Remiander R de MBITS
-- Furthermore sign(X) = sign(R)
--	The correction step is only an adition
---------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use work.mypackage.all;

entity div_nr_integer is
    port (
        X: in STD_LOGIC_VECTOR (NBITS-1 downto 0);
        Y: in STD_LOGIC_VECTOR (MBITS-1 downto 0);
        Q: out STD_LOGIC_VECTOR (NBITS-1 downto 0);
		  R: out STD_LOGIC_VECTOR (MBITS-1 downto 0)
     );
end div_nr_integer;

architecture mult_arch of div_nr_integer is

component complement_cell is
    generic (Length : integer := 8);
    port (
	 		a: in STD_LOGIC_VECTOR (Length-1  downto 0);
			c: in STD_LOGIC;
			ac: out STD_LOGIC_VECTOR (Length-1 downto 0)
     );
end component;

component cond_adder
	port(
		a : in std_logic_vector(MBITS-1 downto 0);
		b : in std_logic_vector(MBITS-1 downto 0);
		x_i : in std_logic;          
		r : out std_logic_vector(MBITS-1 downto 0)
		);
end component;


component nr_cell is
    port (
        a: in STD_LOGIC_VECTOR (MBITS downto 0);
        b: in STD_LOGIC_VECTOR (MBITS-1 downto 0);
        q: in STD_LOGIC;
		  r: out STD_LOGIC_VECTOR (MBITS downto 0)
     );
end component;

type conections is array (0 to NBITS-1) of STD_LOGIC_VECTOR (MBITS downto 0);
Signal wires_in, wires_out: conections;

Signal QQ: STD_LOGIC_VECTOR (NBITS downto 0);
signal adjust, compl_Q, compl_R: STD_LOGIC;
Signal zeros: STD_LOGIC_VECTOR (MBITS-1 downto 0);

Signal PX, PQ:  STD_LOGIC_VECTOR (NBITS-1 downto 0);
Signal PR, PY:  STD_LOGIC_VECTOR (MBITS-1 downto 0);

begin

 positive_x: complement_cell generic map (Length => NBITS)
			    port map( a => X, c => X(NBITS-1), ac => PX);
 positive_y: complement_cell generic map (Length => MBITS)
			    port map( a => Y, c => Y(MBITS-1), ac => PY);


 QQ(NBITS) <=  '0'; --the sign bit	 
 zeros <= (others => '0'); 
 wires_in(0) <= zeros & PX(NBITS-1);  

 divisor: for I in 0 to NBITS-1 generate
    int_mod: nr_cell port map (a => wires_in(I),
  			 b => PY, 
  			 q => QQ(NBITS-I),
  			 r => wires_out(i) );
  end generate;

 wires_conections: for I in 0 to NBITS-2 generate
    QQ(NBITS-i-1) <= wires_out(i)(MBITS);	 
    wires_in(i+1) <= wires_out(i)(MBITS-1 downto 0) & PX(NBITS-I-2);	 
 end generate;

 adjust <= (wires_out(NBITS-1)(MBITS));
 final_adjust: cond_adder port map ( a => wires_out(NBITS-1)(MBITS-1 downto 0),
  			 b => PY, x_i => adjust, r => PR);

 PQ(NBITS-1 downto 1) <= not QQ(NBITS-1 downto 1);
 PQ(0) <= not adjust;

 compl_Q <= X(NBITS-1) xor Y(MBITS-1); 
 compl_R <= X(NBITS-1); 

 complement_Q: complement_cell generic map (Length => NBITS)
			    port map( a => PQ, c => compl_Q, ac => Q);
 complement_R: complement_cell generic map (Length => MBITS)
			    port map( a => PR, c => compl_R, ac => R);


end mult_arch;


