---------------------------------------------------------------
-- NON RESTORING Division algorithm for integer (signed) X and Y natural (unsigned)
-- NBITS >= MBITS
-- Returns a Quotient Q of NBITS and a Remiander R de MBITS
-- Furthermore sign(X) = sign(R)
--	The correction step is only an adition
---------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use work.mypackage.all;

entity div_nr_su is
    port (
        X: in STD_LOGIC_VECTOR (NBITS-1 downto 0);
        Y: in STD_LOGIC_VECTOR (MBITS-1 downto 0);
        Q: out STD_LOGIC_VECTOR (NBITS-1 downto 0);
		  R: out STD_LOGIC_VECTOR (MBITS downto 0)
     );
end div_nr_su;

architecture mult_arch of div_nr_su is

COMPONENT correction_circuit
	PORT(
		rem_not_adj : IN std_logic_vector(MBITS downto 0);
		q_not_adj : in std_logic_vector(NBITS-1 downto 0);
		y : IN std_logic_vector(MBITS-1 downto 0);
		sign_X, sign_R : IN std_logic;          
		q : OUT std_logic_vector(NBITS-1 downto 0);
		r : OUT std_logic_vector(MBITS downto 0)
		);
END COMPONENT;

component nr_cell is
    port (
        a: in STD_LOGIC_VECTOR (MBITS downto 0);
        b: in STD_LOGIC_VECTOR (MBITS-1 downto 0);
        q: in STD_LOGIC;
		  r: out STD_LOGIC_VECTOR (MBITS downto 0)
     );
end component;

type conections is array (0 to NBITS-1) of STD_LOGIC_VECTOR (MBITS downto 0);
Signal wires_in, wires_out: conections;

Signal  QQ: STD_LOGIC_VECTOR (NBITS downto 0);
signal  sign_rem: STD_LOGIC;
Signal  signbits: STD_LOGIC_VECTOR (MBITS-1 downto 0);

Signal  QQ_neg: STD_LOGIC_VECTOR (NBITS-1 downto 0);

begin

 QQ(NBITS) <=  X(NBITS-1); --the sign bit	 
 signbits <= (others => X(NBITS-1)); 
 wires_in(0) <= signbits & X(NBITS-1);  

 divisor: for I in 0 to NBITS-1 generate
    int_mod: nr_cell port map (a => wires_in(I),
  			 b => Y, 
  			 q => QQ(NBITS-I),
  			 r => wires_out(i) );
  end generate;
  			 

 wires_conections: for I in 0 to NBITS-2 generate
    QQ(NBITS-i-1) <= wires_out(i)(MBITS);	 
    wires_in(i+1) <= wires_out(i)(MBITS-1 downto 0) & X(NBITS-I-2);	 
 end generate;

 sign_rem <= (wires_out(NBITS-1)(MBITS));
 QQ(0) <=  '0';	 
 QQ_neg <= not (QQ(NBITS-1 downto 0)) ;

 final_adjust: correction_circuit port map ( rem_not_adj => wires_out(NBITS-1)(MBITS downto 0),
			 Q_not_adj => QQ_neg, 
  			 y => Y, 
			 sign_X => X(NBITS-1),
			 sign_R => sign_rem, 
			 q => Q, r => R);

end mult_arch;


